# -*- coding: utf-8 -*-

__all__ = ["QuadPotentialDenseAdapt", "get_dense_nuts_step", "sample"]

import numpy as np
import pymc3 as pm
import theano
from pymc3.model import all_continuous, modelcontext
from pymc3.step_methods.hmc.quadpotential import QuadPotential
from pymc3.step_methods.step_sizes import DualAverageAdaptation
from scipy.linalg import LinAlgError, cholesky, solve_triangular

from .utils import deprecated, logger


class QuadPotentialDenseAdapt(QuadPotential):
    """Adapt a dense mass matrix from the sample covariances."""

    def __init__(
        self,
        n,
        initial_mean=None,
        initial_cov=None,
        initial_weight=0,
        adaptation_window=101,
        doubling=True,
        update_steps=None,
        dtype="float64",
    ):
        if initial_mean is None:
            initial_mean = np.zeros(n, dtype=dtype)
        if initial_cov is None:
            initial_cov = np.eye(n, dtype=dtype)
            initial_weight = 1

        if initial_cov is not None and initial_cov.ndim != 2:
            raise ValueError("Initial covariance must be two-dimensional.")
        if initial_mean is not None and initial_mean.ndim != 1:
            raise ValueError("Initial mean must be one-dimensional.")
        if initial_cov is not None and initial_cov.shape != (n, n):
            raise ValueError(
                "Wrong shape for initial_cov: expected %s got %s"
                % (n, initial_cov.shape)
            )
        if len(initial_mean) != n:
            raise ValueError(
                "Wrong shape for initial_mean: expected %s got %s"
                % (n, len(initial_mean))
            )

        self.dtype = dtype
        self._n = n
        self._cov = np.array(initial_cov, dtype=self.dtype, copy=True)
        self._cov_theano = theano.shared(self._cov)
        self._chol = cholesky(self._cov, lower=True)
        self._chol_error = None
        self._foreground_cov = _WeightedCovariance(
            self._n, initial_mean, initial_cov, initial_weight, self.dtype
        )
        self._background_cov = _WeightedCovariance(self._n, dtype=self.dtype)
        self._n_samples = 0

        # For backwards compatibility
        self._doubling = doubling
        self._adaptation_window = int(adaptation_window)
        self._previous_update = 0

        # New interface
        if update_steps is None:
            self._update_steps = None
        else:
            self._update_steps = np.atleast_1d(update_steps).astype(int)

    def velocity(self, x, out=None):
        return np.dot(self._cov, x, out=out)

    def energy(self, x, velocity=None):
        if velocity is None:
            velocity = self.velocity(x)
        return 0.5 * np.dot(x, velocity)

    def velocity_energy(self, x, v_out):
        self.velocity(x, out=v_out)
        return self.energy(x, v_out)

    def random(self):
        vals = np.random.normal(size=self._n).astype(self.dtype)
        return solve_triangular(self._chol.T, vals, overwrite_b=True)

    def _update_from_weightvar(self, weightvar):
        weightvar.current_covariance(out=self._cov)
        try:
            self._chol = cholesky(self._cov, lower=True)
        except (LinAlgError, ValueError) as error:
            self._chol_error = error
        self._cov_theano.set_value(self._cov)

    def update(self, sample, grad, tune):
        if not tune:
            return

        self._foreground_cov.add_sample(sample, weight=1)
        self._background_cov.add_sample(sample, weight=1)
        self._update_from_weightvar(self._foreground_cov)

        # Support the two methods for updating the mass matrix
        delta = self._n_samples - self._previous_update
        do_update = (
            self._update_steps is not None
            and self._n_samples in self._update_steps
        ) or (self._update_steps is None and delta >= self._adaptation_window)
        if do_update:
            self._foreground_cov = self._background_cov
            self._background_cov = _WeightedCovariance(
                self._n, dtype=self.dtype
            )

            if self._update_steps is None:
                self._previous_update = self._n_samples
                if self._doubling:
                    self._adaptation_window *= 2

        self._n_samples += 1

    def raise_ok(self, vmap):
        if self._chol_error is not None:
            raise ValueError("{0}".format(self._chol_error))


class _WeightedCovariance:
    """Online algorithm for computing mean and covariance."""

    def __init__(
        self,
        nelem,
        initial_mean=None,
        initial_covariance=None,
        initial_weight=0,
        dtype="float64",
    ):
        self._dtype = dtype
        self.n_samples = float(initial_weight)
        if initial_mean is None:
            self.mean = np.zeros(nelem, dtype=dtype)
        else:
            self.mean = np.array(initial_mean, dtype=dtype, copy=True)
        if initial_covariance is None:
            self.raw_cov = np.eye(nelem, dtype=dtype)
        else:
            self.raw_cov = np.array(initial_covariance, dtype=dtype, copy=True)

        self.raw_cov[:] *= self.n_samples

        if self.raw_cov.shape != (nelem, nelem):
            raise ValueError("Invalid shape for initial covariance.")
        if self.mean.shape != (nelem,):
            raise ValueError("Invalid shape for initial mean.")

    def add_sample(self, x, weight):
        x = np.asarray(x)
        self.n_samples += 1
        old_diff = x - self.mean
        self.mean[:] += old_diff / self.n_samples
        new_diff = x - self.mean
        self.raw_cov[:] += weight * new_diff[:, None] * old_diff[None, :]

    def current_covariance(self, out=None):
        if self.n_samples == 0:
            raise ValueError("Can not compute covariance without samples.")
        if out is not None:
            return np.divide(self.raw_cov, self.n_samples - 1, out=out)
        else:
            return (self.raw_cov / (self.n_samples - 1)).astype(self._dtype)

    def current_mean(self):
        return np.array(self.mean, dtype=self._dtype)


class WindowedDualAverageAdaptation(DualAverageAdaptation):
    def __init__(self, update_steps, initial_step, target, *args, **kwargs):
        self.update_steps = np.atleast_1d(update_steps).astype(int)
        self.targets = np.atleast_1d(target) + np.zeros_like(self.update_steps)
        super(WindowedDualAverageAdaptation, self).__init__(
            initial_step, self.targets[0], *args, **kwargs
        )
        self._initial_step = initial_step
        self._n_samples = 0

    def reset(self):
        self._hbar = 0.0
        self._log_step = np.log(self._initial_step)
        self._log_bar = self._log_step
        self._mu = np.log(10 * self._initial_step)
        self._count = 1
        self._tuned_stats = []

    def update(self, accept_stat, tune):
        if self._n_samples in self.update_steps:
            self._target = float(
                self.targets[np.where(self.update_steps == self._n_samples)]
            )
            self._n_samples += 1
            self.reset()
            return

        self._n_samples += 1
        super(WindowedDualAverageAdaptation, self).update(accept_stat, tune)


def build_schedule(
    tune, warmup_window=50, adapt_window=50, cooldown_window=50
):
    if warmup_window + adapt_window + cooldown_window > tune:
        logger.warn(
            "there are not enough tuning steps to accomodate the tuning "
            "schedule; assigning automatically as 20%/70%/10%"
        )
        warmup_window = np.ceil(0.2 * tune).astype(int)
        cooldown_window = np.ceil(0.1 * tune).astype(int)
        adapt_window = tune - warmup_window - cooldown_window

    t = warmup_window
    delta = adapt_window
    update_steps = []
    while t < tune - cooldown_window:
        t += delta
        delta = 2 * delta
        if t + delta > tune - cooldown_window:
            update_steps.append(tune - cooldown_window)
            break
        update_steps.append(t)

    update_steps = np.array(update_steps, dtype=int)
    if np.any(update_steps) <= 0:
        raise ValueError("invalid tuning schedule")

    return np.append(warmup_window, update_steps)


@deprecated("the sample function from the pymc3-ext library")
def sample(
    *,
    draws=1000,
    tune=1000,
    model=None,
    step_kwargs=None,
    warmup_window=50,
    adapt_window=50,
    cooldown_window=100,
    initial_accept=None,
    target_accept=0.9,
    gamma=0.05,
    k=0.75,
    t0=10,
    **kwargs,
):
    # Check that we're in a model context and that all the variables are
    # continuous
    model = modelcontext(model)
    if not all_continuous(model.vars):
        raise ValueError(
            "NUTS can only be used for models with only "
            "continuous variables."
        )
    start = kwargs.get("start", None)
    if start is None:
        start = model.test_point
    mean = model.dict_to_array(start)

    update_steps = build_schedule(
        tune,
        warmup_window=warmup_window,
        adapt_window=adapt_window,
        cooldown_window=cooldown_window,
    )

    potential = QuadPotentialDenseAdapt(
        model.ndim,
        initial_mean=mean,
        initial_weight=10,
        update_steps=update_steps,
    )

    if "step" in kwargs:
        step = kwargs["step"]
    else:
        if step_kwargs is None:
            step_kwargs = {}
        step = pm.NUTS(
            potential=potential,
            model=model,
            target_accept=target_accept,
            **step_kwargs,
        )

    if "target_accept" in step_kwargs and target_accept is not None:
        raise ValueError(
            "'target_accept' cannot be given as a keyword argument and in "
            "'step_kwargs'"
        )
    target_accept = step_kwargs.pop("target_accept", target_accept)
    if initial_accept is None:
        target = target_accept
    else:
        if initial_accept > target_accept:
            raise ValueError(
                "initial_accept must be less than or equal to target_accept"
            )
        target = initial_accept + (target_accept - initial_accept) * np.sqrt(
            np.arange(len(update_steps)) / (len(update_steps) - 1)
        )
    step.step_adapt = WindowedDualAverageAdaptation(
        update_steps, step.step_size, target, gamma, k, t0
    )

    kwargs["step"] = step
    return pm.sample(draws=draws, tune=tune, model=model, **kwargs)


@deprecated("the init='full_adapt' argument to pm.sample")
def get_dense_nuts_step(
    start=None,
    adaptation_window=101,
    doubling=True,
    initial_weight=10,
    use_hessian=False,
    use_hessian_diag=False,
    hessian_regularization=1e-8,
    model=None,
    **kwargs,
):
    """Get a NUTS step function with a dense mass matrix

    The entries in the mass matrix will be tuned based on the sample
    covariances during tuning. All extra arguments are passed directly to
    ``pymc3.NUTS``.

    Args:
        start (dict, optional): A starting point in parameter space. If not
            provided, the model's ``test_point`` is used.
        adaptation_window (int, optional): The (initial) size of the window
            used for sample covariance estimation.
        doubling (bool, optional): If ``True`` (default) the adaptation window
            is doubled each time the matrix is updated.

    """
    model = modelcontext(model)

    if not all_continuous(model.vars):
        raise ValueError(
            "NUTS can only be used for models with only "
            "continuous variables."
        )

    if start is None:
        start = model.test_point
    mean = model.dict_to_array(start)

    if use_hessian or use_hessian_diag:
        try:
            import numdifftools as nd
        except ImportError:
            raise ImportError(
                "The 'numdifftools' package is required for Hessian "
                "computations"
            )

        logger.info("Numerically estimating Hessian matrix")
        if use_hessian_diag:
            hess = nd.Hessdiag(model.logp_array)(mean)
            var = np.diag(-1.0 / hess)
        else:
            hess = nd.Hessian(model.logp_array)(mean)
            var = -np.linalg.inv(hess)

        factor = 1
        success = False
        while not success:
            var[np.diag_indices_from(var)] += factor * hessian_regularization

            try:
                np.linalg.cholesky(var)
            except np.linalg.LinAlgError:
                factor *= 2
            else:
                success = True

    else:
        var = np.eye(len(mean))

    potential = QuadPotentialDenseAdapt(
        model.ndim,
        initial_mean=mean,
        initial_cov=var,
        initial_weight=initial_weight,
        adaptation_window=adaptation_window,
        doubling=doubling,
    )

    return pm.NUTS(potential=potential, model=model, **kwargs)



# -*- coding: utf-8 -*-

__all__ = ["get_cl", "limbdark"]

import numpy as np
import theano
import theano.tensor as tt
from scipy.optimize import bisect

from ..utils import as_tensor_variable
from . import driver
from .helpers import resize_or_set


class GetClRev(theano.Op):
    __props__ = ()
    itypes = (tt.dvector,)
    otypes = (tt.dvector,)

    def infer_shape(self, node, shapes):
        return shapes

    def perform(self, node, inputs, outputs):
        bc = inputs[0]
        bu = resize_or_set(outputs, 0, bc.shape)
        driver.get_cl_rev(bc, bu)


get_cl_rev = GetClRev()


class GetCl(theano.Op):

    __props__ = ()
    itypes = (tt.dvector,)
    otypes = (tt.dvector,)

    def infer_shape(self, node, shapes):
        return shapes

    def perform(self, node, inputs, outputs):
        u = inputs[0]
        c = resize_or_set(outputs, 0, u.shape)
        driver.get_cl(u, c)

    def grad(self, inputs, gradients):
        return (get_cl_rev(gradients[0]),)

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)


get_cl = GetCl()


class LimbDark(theano.Op):
    __props__ = ()

    def __init__(self):
        self.ld = driver.LimbDark()
        super().__init__()

    def __getstate__(self):
        return {}

    def __setstate__(self, data):
        self.ld = driver.LimbDark()

    def make_node(self, *inputs):
        in_args = [as_tensor_variable(i) for i in inputs]
        if any(i.dtype != "float64" for i in in_args):
            raise ValueError(
                "float64 dtypes are required for LimbDark op; "
                "got:\n{0}".format([i.dtype for i in inputs])
            )
        out_args = [
            in_args[1].type(),
            tt.TensorType(
                dtype="float64", broadcastable=[False] * (in_args[1].ndim + 1)
            )(),
            in_args[1].type(),
            in_args[2].type(),
        ]
        return theano.Apply(self, in_args, out_args)

    def infer_shape(self, node, shapes):
        return (
            shapes[1],
            list(shapes[0]) + list(shapes[1]),
            shapes[1],
            shapes[2],
        )

    def perform(self, node, inputs, outputs):
        cl, b, r, los = inputs
        f = resize_or_set(outputs, 0, b.shape)
        dfdcl = resize_or_set(outputs, 1, cl.shape + b.shape)
        dfdb = resize_or_set(outputs, 2, b.shape)
        dfdr = resize_or_set(outputs, 3, b.shape)
        self.ld.apply(cl, b, r, los, f, dfdcl, dfdb, dfdr)

    def grad(self, inputs, gradients):
        c, b, r, los = inputs
        f, dfdcl, dfdb, dfdr = self(*inputs)
        bf = gradients[0]
        for i, g in enumerate(gradients[1:]):
            if not isinstance(g.type, theano.gradient.DisconnectedType):
                raise ValueError(
                    "can't propagate gradients wrt parameter {0}".format(i + 1)
                )
        bc = tt.sum(
            tt.reshape(bf, (1, bf.size))
            * tt.reshape(dfdcl, (c.size, bf.size)),
            axis=-1,
        )
        bb = bf * dfdb
        br = bf * dfdr
        return bc, bb, br, tt.zeros_like(los)

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)


limbdark = LimbDark()


#
# Helpers for Radius From Occ Area
#


def kite_area(r, b):
    abc = tt.sort(tt.stack((r, b, tt.ones_like(r)), axis=0), axis=0)
    A = abc[0]
    B = abc[1]
    C = abc[2]
    return 0.5 * tt.sqrt(
        tt.abs_((A + (B + C)) * (C - (A - B)) * (C + (A - B)) * (A + (B - C)))
    )


def depth(r, b):
    # depth = 1 - s0 / pi; where s0 is from Agol+
    b = tt.abs_(b)
    r = tt.abs_(r)
    b2 = b ** 2
    r2 = r ** 2
    opr = 1 + r
    omr = 1 - r
    rmo = r - 1

    # Case 2
    a = kite_area(r, b)
    twoa = 2 * a
    k0 = tt.arctan2(twoa, rmo * opr + b2)
    k1 = tt.arctan2(twoa, omr * opr + b2)
    case2 = (k1 + r2 * k0 - a) / np.pi

    return tt.switch(
        tt.le(opr, b),
        tt.zeros_like(r),
        tt.switch(
            tt.and_(tt.lt(tt.abs_(omr), b), tt.lt(b, opr)),
            case2,
            tt.switch(tt.le(b, omr), r2, tt.ones_like(r)),
        ),
    )


def depth_grad(r, b):
    # depth = 1 - s0 / pi; where s0 is from Agol+
    b = tt.abs_(b)
    r = tt.abs_(r)
    b2 = b ** 2
    opr = 1 + r
    omr = 1 - r
    rmo = r - 1

    # Case 2
    a = kite_area(r, b)
    twor = 2 * r
    twoa = 2 * a
    k0 = tt.arctan2(twoa, rmo * opr + b2)
    dr = twor * k0 / np.pi
    db = -twoa / (b * np.pi)

    zero = tt.zeros_like(r)
    return (
        tt.switch(
            tt.le(opr, b),
            zero,
            tt.switch(
                tt.and_(tt.lt(tt.abs_(omr), b), tt.lt(b, opr)),
                dr,
                tt.switch(tt.le(b, omr), twor, zero),
            ),
        ),
        tt.switch(
            tt.le(opr, b),
            zero,
            tt.switch(
                tt.and_(tt.lt(tt.abs_(omr), b), tt.lt(b, opr)), db, zero
            ),
        ),
    )


#
# /helpers
#


class RadiusFromOccArea(theano.Op):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

        # Set up the problem using Theano ops
        b = tt.dscalar()
        delta = tt.dscalar()
        r = tt.dscalar()

        # We'll need test vals to keep PyMC3 happy
        b.tag.test_value = 0.0
        delta.tag.test_value = 0.0
        r.tag.test_value = 0.0

        # Compile the objective function (this is all a little meta)
        value = delta - depth(r, b)
        self.func = theano.function([r, b, delta], value)

        super().__init__()

    def make_node(self, delta, b):
        in_args = [as_tensor_variable(delta), as_tensor_variable(b)]
        out_args = [in_args[1].type()]
        return theano.Apply(self, in_args, out_args)

    def infer_shape(self, node, shapes):
        return (shapes[1],)

    def perform(self, node, inputs, outputs):
        delta, b = inputs
        shape = np.shape(b)

        b = np.atleast_1d(b)
        delta = np.atleast_1d(delta)
        if b.ndim != 1 or b.shape != delta.shape:
            raise ValueError("dimension mismatch")

        if np.any(b < 0) or np.any(delta < 0) or np.any(delta > 1):
            raise ValueError(
                "invalid input; valid ranges 0 <= delta <= 1; 0 <= b"
            )

        # Compute the bounds
        rmin = np.maximum(b - 1, 0)
        rmax = b + 1

        # Solve for each entry
        r = np.zeros_like(b)
        for n in range(len(b)):
            args = (b[n], delta[n])
            r[n] = bisect(
                self.func, rmin[n], rmax[n], args=args, **self.kwargs
            )

        outputs[0][0] = r.reshape(shape)

    def grad(self, inputs, gradients):
        delta, b = inputs
        r = self(*inputs)
        br = gradients[0]

        dr, db = depth_grad(r, b)
        bdelta = br / dr
        bb = -db * bdelta

        return bdelta, bb

    def R_op(self, inputs, eval_points):
        if eval_points[0] is None:
            return eval_points
        return self.grad(inputs, eval_points)



# -*- coding: utf-8 -*-

__all__ = ["UnitUniform", "UnitVector", "UnitDisk", "Angle", "Periodic"]

from pymc3_ext.distributions import (
    Angle,
    Periodic,
    UnitDisk,
    UnitUniform,
    UnitVector,
)



# -*- coding: utf-8 -*-

import numpy as np
import pytest
import theano
import theano.tensor as tt

from exoplanet.theano_ops.rebound import ReboundOp
from exoplanet.theano_ops.test_tools import InferShapeTester

try:
    import rebound  # NOQA
except ImportError:
    pytest.skip(
        "skipping rebound tests when not installed", allow_module_level=True
    )


@pytest.mark.filterwarnings("ignore:For better performance")
class TestRebound(InferShapeTester):
    def setup_method(self):
        super().setup_method()
        self.op_class = ReboundOp
        self.op = ReboundOp()

    def get_args(self):
        m_val = np.array([1.3, 1e-3, 1e-5])
        x_val = np.zeros((3, 6))
        x_val[1, 0] = 15.0
        x_val[1, 4] = 0.4
        x_val[2, 0] = 100.0
        x_val[2, 4] = 0.2
        t = np.linspace(100, 1000, 12)

        m = tt.dvector()
        x = tt.dmatrix()

        f = theano.function([m, x], self.op(m, x, t)[0])

        return t, f, [m, x], [m_val, x_val]

    def test_basic(self):
        _, f, _, in_args = self.get_args()
        f(*in_args)

    def test_infer_shape(self):
        t, f, args, arg_vals = self.get_args()
        self._compile_and_check(
            args, self.op(*(list(args) + [t])), arg_vals, self.op_class
        )

    def test_grad(self):
        t, _, _, in_args = self.get_args()
        func = lambda *args: self.op(*(list(args) + [t]))[0]  # NOQA
        tt.verify_grad(func, in_args, n_tests=1, rng=np.random)



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from exoplanet.distributions.deprecated import (
    RadiusImpact,
    get_joint_radius_impact,
)

from .base_test import _Base


class TestPhyscial(_Base):
    random_seed = 20190919

    def test_radius_impact(self):
        min_radius = 0.01
        max_radius = 0.1
        with self._model():
            with pytest.warns(DeprecationWarning):
                dist = RadiusImpact(
                    "rb", min_radius=min_radius, max_radius=max_radius
                )

            # Test random sampling
            samples = dist.random(size=100)
            assert np.shape(samples) == (100, 2)
            assert np.all(
                (min_radius <= samples[:, 0]) & (samples[:, 0] <= max_radius)
            )

            logp = (
                RadiusImpact.dist(min_radius=min_radius, max_radius=max_radius)
                .logp(samples)
                .eval()
                .flatten()
            )
            assert np.all(np.isfinite(logp))
            assert np.allclose(logp[0], logp)

            trace = self._sample()

        r = trace["rb"][:, 0]
        b = trace["rb"][:, 1]

        # Make sure that the physical constraints are satisfied
        assert np.all((r <= max_radius) & (min_radius <= r))
        assert np.all((b >= 0) & (b <= 1 + r))

    def test_get_joint_radius_impact(self):
        min_radius = 0.01
        max_radius = 0.1
        with self._model() as model:
            with pytest.warns(DeprecationWarning):
                r, b = get_joint_radius_impact(
                    min_radius=min_radius, max_radius=max_radius
                )
            assert model.r is r
            assert model.b is b
            trace = self._sample()

        r = trace["r"]
        b = trace["b"]

        # Make sure that the physical constraints are satisfied
        assert np.all((r <= max_radius) & (min_radius <= r))
        assert np.all((b >= 0) & (b <= 1 + r))



# -*- coding: utf-8 -*-

__all__ = ["contact_points"]

import numpy as np
import theano
import theano.tensor as tt

from ..utils import as_tensor_variable
from . import driver
from .helpers import resize_or_set


class ContactPoints(theano.Op):
    __props__ = ()

    def __init__(self, tol=1e-10):
        self.tol = float(tol)
        super().__init__()

    def make_node(self, *inputs):
        in_args = [as_tensor_variable(i) for i in inputs]
        if any(i.dtype != "float64" for i in in_args):
            raise ValueError(
                "float64 dtypes are required for ContactPoints op; "
                "got:\n{0}".format([i.dtype for i in inputs])
            )
        out_args = [
            in_args[0].type(),
            in_args[0].type(),
            tt.TensorType(
                dtype="int32", broadcastable=[False] * in_args[0].ndim
            )(),
        ]
        return theano.Apply(self, in_args, out_args)

    def infer_shape(self, node, shapes):
        return shapes[0], shapes[0], shapes[0]

    def perform(self, node, inputs, outputs):
        a, e, cosw, sinw, cosi, sini, L = inputs
        M_left = resize_or_set(outputs, 0, a.shape)
        M_right = resize_or_set(outputs, 1, a.shape)
        flag = resize_or_set(outputs, 2, a.shape, dtype=np.int32)
        driver.contact_points(
            a, e, cosw, sinw, cosi, sini, L, M_left, M_right, flag, self.tol
        )


contact_points = ContactPoints()



import numpy as np
import pymc3 as pm
import pytest
import theano.tensor as tt

from exoplanet import optim as op
from exoplanet.optim import optimize

try:
    import torch
except ImportError:
    torch = None


def test_optimize(seed=1234):
    np.random.seed(seed)
    x_val = np.random.randn(5, 3)
    with pm.Model():
        pm.Normal("x", shape=x_val.shape, testval=x_val)
        soln1 = optimize(verbose=False)
        soln2, info = optimize(soln1, return_info=True, verbose=False)

    assert np.allclose(soln1["x"], 0.0)
    assert np.allclose(soln2["x"], 0.0)
    assert info.success


def test_optimize_exception(capsys):
    with pm.Model():
        cov = pm.Normal("cov", mu=np.eye(5), shape=(5, 5))
        chol = tt.slinalg.Cholesky(on_error="raise")(cov)
        pm.MvNormal("x", mu=np.zeros(5), chol=chol, shape=5)
        with pytest.raises(np.linalg.LinAlgError):
            optimize({"cov": np.zeros((5, 5))}, verbose=False)
        captured = capsys.readouterr()
        assert "array:" in captured.out
        assert "point:" in captured.out


def rosenbrock(x):
    return (1 - x[0]) ** 2 + 100 * (x[1] - x[0] ** 2) ** 2


@pytest.mark.skipif(torch is None, reason="torch is not installed")
@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"lr": 1e-4},
        {"lr": 1e-4, "betas": [0.92, 0.96]},
        {"lr": 1e-4, "betas": [0.92, 0.96], "eps": 1e-3},
        {"lr": 1e-4, "weight_decay": 0.1},
        {"amsgrad": True},
    ],
)
def test_adam(kwargs, seed=20200520):
    np.random.seed(seed)
    x0 = np.random.randn(2)

    x_torch = torch.tensor(x0, dtype=torch.float64, requires_grad=True)
    optimizer = torch.optim.Adam([x_torch], **kwargs)

    with pm.Model():
        x = pm.Flat("x", shape=2, testval=x0)
        pm.Potential("rosenbrock", -rosenbrock(x))
        for obj, point in op.optimize_iterator(
            op.Adam(**kwargs), 100, vars=[x]
        ):
            optimizer.zero_grad()
            loss = rosenbrock(x_torch)
            loss.backward()
            optimizer.step()
            assert np.allclose(x_torch.detach().numpy(), point["x"])



#!/usr/bin/env python

# Inspired by:
# https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/

import codecs
import os
import re
import sys

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

# PROJECT SPECIFIC

NAME = "exoplanet"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", "exoplanet", "__init__.py")
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
]
SETUP_REQUIRES = ["setuptools>=40.6.0", "setuptools_scm"]
INSTALL_REQUIRES = [
    "pybind11>=2.4",
    "numpy>=1.13.0",
    "pymc3>=3.5",
    "astropy>=3.1",
    "pymc3-ext>=0.0.1",
]
EXTRA_REQUIRE = {
    "test": [
        "scipy",
        "nose",
        "parameterized",
        "arviz",
        "pytest",
        "pytest-cov>=2.6.1",
        "pytest-env",
        "coveralls",
        "pybind11",
        "batman-package",
        "rebound; sys_platform != 'win32'",
    ],
    "docs": [
        "sphinx>=1.7.5",
        "pandoc",
        "jupyter",
        "ipywidgets",
        "sphinx-typlog-theme",
        "nbformat",
        "nbconvert",
        "corner",
        "lightkurve",
        "jupytext",
        "rtds_action",
        "nbsphinx",
    ],
    "tutorials": [
        "jupytext<1.7.0",
        "jupyter",
        "nbconvert",
        "matplotlib",
        "corner",
        "lightkurve",
    ],
    "nbody": [
        "rebound; sys_platform != 'win32'",
        "rebound_pymc3>=0.0.3; sys_platform != 'win32'",
    ],
}
EXTRA_REQUIRE["dev"] = (
    EXTRA_REQUIRE["test"]
    + EXTRA_REQUIRE["docs"]
    + EXTRA_REQUIRE["nbody"]
    + [
        "pre-commit",
        "black",
        "black_nbconvert",
        "isort",
        "toml",
        "flake8",
        "nbstripout",
        "jupytext",
        "radvel",
        "jupyterlab",
        "lightkurve",
        "pep517",
        "twine",
    ]
)

# END PROJECT SPECIFIC

# PYBIND11


class get_pybind_include:
    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11

        return pybind11.get_include(self.user)


class get_numpy_include:
    def __str__(self):
        import numpy

        return numpy.get_include()


class custom_build_ext(build_ext):
    c_opts = {"msvc": ["/EHsc"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    if sys.platform == "darwin":
        darwin_opts = [
            "-stdlib=libc++",
            "-mmacosx-version-min=10.14",
            "-march=native",
        ]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def has_flag(self, flagname):
        import tempfile

        import setuptools

        with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
            f.write("int main (int argc, char **argv) { return 0; }")
            try:
                self.compiler.compile([f.name], extra_postargs=[flagname])
            except setuptools.distutils.errors.CompileError:
                return False
        return True

    def cpp_flag(self):
        flags = ["-std=c++17", "-std=c++14", "-std=c++11"]

        for flag in flags:
            if self.has_flag(flag):
                return flag

        raise RuntimeError(
            "Unsupported compiler. At least C++11 support is needed."
        )

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append(
                '-DVERSION_INFO="%s"' % self.distribution.get_version()
            )
            opts.append(self.cpp_flag())
            if self.has_flag("-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append(
                '/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version()
            )
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


include_dirs = [
    "src/exoplanet/theano_ops/lib/include",
    "src/exoplanet/theano_ops/lib/vendor/eigen",
    get_numpy_include(),
    get_pybind_include(),
    get_pybind_include(user=True),
]
ext_modules = [
    Extension(
        "exoplanet.theano_ops.driver",
        ["src/exoplanet/theano_ops/driver.cpp"],
        include_dirs=include_dirs,
        language="c++",
    )
]

# END PYBIND11

HERE = os.path.dirname(os.path.realpath(__file__))


def read(*parts):
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta, meta_file=read(META_PATH)):
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), meta_file, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


if __name__ == "__main__":
    setup(
        name=NAME,
        use_scm_version={
            "write_to": os.path.join(
                "src", NAME, "{0}_version.py".format(NAME)
            ),
            "write_to_template": '__version__ = "{version}"\n',
        },
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        url=find_meta("uri"),
        license=find_meta("license"),
        description=find_meta("description"),
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        packages=PACKAGES,
        package_dir={"": "src"},
        include_package_data=True,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRA_REQUIRE,
        classifiers=CLASSIFIERS,
        setup_requires=SETUP_REQUIRES,
        zip_safe=False,
        ext_modules=ext_modules,
        cmdclass={"build_ext": custom_build_ext},
    )



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.6.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# + nbsphinx="hidden"
# %matplotlib inline
# -

# + nbsphinx="hidden"
# %run notebook_setup
# -

# # Astrometric fitting

# In this tutorial we'll walk through the simplest astrometric example with `exoplanet` and then explain how to build up a more complicated example with parallax measurements. For our dataset, we'll use astrometric and radial velocity observations of a binary star system.
#
# Astrometric observations usually consist of measurements of the separation and position angle of the secondary star (or directly imaged exoplanet), relative to the primary star as a function of time. The simplest astrometric orbit (in terms of number of parameters), describes the orbit using a semi-major axis `a_ang` measured in *arcseconds*, since the distance to the system is assumed to be unknown. We'll work through this example first, then introduce the extra constraints provided by parallax information.
#
# ## Data
# First, let's load and examine the data. We'll use the astrometric measurements of HR 466 (HD 10009) as compiled by [Pourbaix 1998](https://ui.adsabs.harvard.edu/#abs/1998A&AS..131..377P/abstract). The speckle observations are originally from [Hartkopf et al. 1996](https://ui.adsabs.harvard.edu/#abs/1996AJ....111..370H/abstract).

# +
from astropy.io import ascii
from astropy.time import Time

# grab the formatted data and do some munging
dirname = "https://gist.github.com/iancze/262aba2429cb9aee3fd5b5e1a4582d4d/raw/c5fa5bc39fec90d2cc2e736eed479099e3e598e3/"

astro_data_full = ascii.read(
    dirname + "astro.txt", format="csv", fill_values=[(".", "0")]
)

# convert UT date to JD
astro_dates = Time(astro_data_full["date"].data, format="decimalyear")

# Following the Pourbaix et al. 1998 analysis, we'll limit ourselves to the highest quality data
# since the raw collection of data outside of these ranges has some ambiguities in swapping
# the primary and secondary star
ind = (
    (astro_dates.value > 1975.0)
    & (astro_dates.value < 1999.73)
    & (~astro_data_full["rho"].mask)
    & (~astro_data_full["PA"].mask)
)

astro_data = astro_data_full[ind]

astro_yrs = astro_data["date"]
astro_dates.format = "jd"
astro_jds = astro_dates[ind].value
# -

# Many of these measurements in this heterogeneous dataset do not have reported error measurements. For these, we assume a modest uncertainty of $1^\circ$ in position angle and $0.01^{\prime\prime}$ in separation for the sake of specifying something, but we'll include a jitter term for both of these measurements as well. The scatter in points around the final solution will be a decent guide of what the measurement uncertainties actually were.

# +
import numpy as np

astro_data["rho_err"][astro_data["rho_err"].mask == True] = 0.01
astro_data["PA_err"][astro_data["PA_err"].mask == True] = 1.0

# Convert all masked frames to be raw np arrays, since theano has issues with astropy masked columns
rho_data = np.ascontiguousarray(astro_data["rho"], dtype=float)  # arcsec
rho_err = np.ascontiguousarray(astro_data["rho_err"], dtype=float)

# The position angle measurements come in degrees in the range [0, 360].
# We'll convert this to radians in the range [-pi, pi]
deg = np.pi / 180.0
theta_data = np.ascontiguousarray(astro_data["PA"] * deg, dtype=float)
theta_data[theta_data > np.pi] -= 2 * np.pi

theta_err = np.ascontiguousarray(astro_data["PA_err"] * deg)  # radians
# -

# ## Astrometric conventions
#
# The conventions describing the orientation of the orbits are described in detail in the *exoplanet* paper; we summarize them briefly here. Generally, we follow the conventions from Pourbaix et al. 1998, which are a consistent set conforming to the right-hand-rule and the conventions of the visual binary field, where the ascending node is that where the secondary is *receeding* from the observer (without radial velocity information, there is a $\pi$ degeneracy in which node is ascending, and so common practice in the literature is to report a value in the range $[0,\pi]$). The orbital inclination ranges from $[0, \pi$]. $i = 0$ describes a face-on orbit rotating counter-clockwise on the sky plane, while $i=\pi$ describes a face-on orbit rotating clockwise on the sky. $i = \pi/2$ is an edge-on orbit.
#
# The observer frame $X$, $Y$, $Z$ is oriented on the sky such that $+Z$ points towards the observer, $X$ is the north axis, and $Y$ is the east axis. *All* angles are measured in radians, and the position angle is returned in the range $[-\pi, \pi]$, which is the degrees east of north (be sure to check your data is in this format too!) The radial velocity is still defined such that a positive radial velocity corresponds to motion away from the observer.
#
# In an astrometric-only orbit, it is common practice in the field to report $\omega = \omega_\mathrm{secondary}$, whereas with an RV orbit it is generally common practice to report $\omega = \omega_\mathrm{primary}$. The result is that unless the authors specify what they're using, in a joint astrometric-RV orbit there is an ambiguity to which $\omega$ the authors mean, since $\omega_\mathrm{primary} = \omega_\mathrm{secondary} + \pi$. To standardize this across the *exoplanet* package, in all orbits (including astrometric-only) $\omega = \omega_\mathrm{primary}$.

# +
import matplotlib.pyplot as plt

# Make a plot of the astrometric data on the sky
# The convention is that North is up and East is left
fig, ax = plt.subplots(nrows=1, figsize=(4, 4))

xs = rho_data * np.cos(theta_data)  # X is north
ys = rho_data * np.sin(theta_data)  # Y is east
ax.plot(ys, xs, ".k")
ax.set_ylabel(r"$\Delta \delta$ ['']")
ax.set_xlabel(r"$\Delta \alpha \cos \delta$ ['']")
ax.invert_xaxis()
ax.plot(0, 0, "k*")
ax.set_aspect("equal", "datalim")
# -

# The plot on the sky is helpful to look at, but the "raw" measurements are the values of $\rho$ (separation) and $\theta$ (also called P.A., position angle) that we listed in our data table, and that the measurement uncertainties live on these values as nice Gaussians. So, to visualize this space more clearly, we can plot $\rho$ vs. time and P.A. vs. time.

# +
fig, ax = plt.subplots(nrows=2, sharex=True)
ax[0].errorbar(astro_yrs, rho_data, yerr=rho_err, fmt=".k", lw=1, ms=5)
ax[0].set_ylabel(r'$\rho\,$ ["]')

ax[1].errorbar(astro_yrs, theta_data, yerr=theta_err, fmt=".k", lw=1, ms=5)
ax[1].set_ylabel(r"P.A. [radians]")
_ = ax[1].set_xlabel("time [years]")
# -

# ## Fitting the astrometric orbit with *exoplanet*
#
# To get started, let's import the relative packages from *exoplanet*, plot up a preliminary orbit from the literature, and then sample to find the best parameters.

# + active=""
# .. note:: Orbits in *exoplanet* generally specify the semi-major axis in units of solar radii `R_sun`. For transits and RV orbits, you usually have enough external information (e.g., estimate of stellar mass from spectral type) to put a physical scale onto the orbit. For the most basic of astrometric orbits without parallax information, however, this information can be lacking and thus it makes sense to fit for the semi-major axis in units of `arcseconds`. But, `exoplanet` is modeling a real orbit (where semi-major axis is in units of `R_sun`), so we do need to at least provide a fake parallax to convert from arcseconds to `R_sun.`

# +
import pymc3 as pm
import theano
import theano.tensor as tt

import exoplanet as xo
from exoplanet.distributions import Angle

from astropy import constants

# conversion constant from au to R_sun
au_to_R_sun = (constants.au / constants.R_sun).value

# Just to get started, let's take a look at the orbit using the best-fit parameters from Pourbaix et al. 1998

# Orbital elements from Pourbaix et al. 1998
# For the relative astrometric fit, we only need the following parameters
a_ang = 0.324  # arcsec
parallax = 1  # arcsec (meaningless choice for now)
a = a_ang * au_to_R_sun / parallax
e = 0.798
i = 96.0 * deg  # [rad]
omega = 251.6 * deg - np.pi  # Pourbaix reports omega_2, but we want omega_1
Omega = 159.6 * deg
P = 28.8 * 365.25  # days

T0 = Time(1989.92, format="decimalyear")
T0.format = "jd"
T0 = T0.value  # [Julian Date]

# instantiate the orbit
orbit = xo.orbits.KeplerianOrbit(
    a=a, t_periastron=T0, period=P, incl=i, ecc=e, omega=omega, Omega=Omega
)

# The position functions take an optional argument parallax to convert from
# physical units back to arcseconds
t = np.linspace(T0 - P, T0 + P, num=200)  # days
rho, theta = theano.function([], orbit.get_relative_angles(t, parallax))()

# Plot the orbit
fig, ax = plt.subplots(nrows=1, figsize=(4, 4))

xs = rho * np.cos(theta)  # X is north
ys = rho * np.sin(theta)  # Y is east
ax.plot(ys, xs, color="C0", lw=1)

# plot the data
xs = rho_data * np.cos(theta_data)  # X is north
ys = rho_data * np.sin(theta_data)  # Y is east
ax.plot(ys, xs, ".k")

ax.set_ylabel(r"$\Delta \delta$ ['']")
ax.set_xlabel(r"$\Delta \alpha \cos \delta$ ['']")
ax.invert_xaxis()
ax.plot(0, 0, "k*")
ax.set_aspect("equal", "datalim")
ax.set_title("initial orbit")

fig, ax = plt.subplots(nrows=2, sharex=True, figsize=(6, 6))
ax[0].errorbar(astro_jds, rho_data, yerr=rho_err, fmt=".k", lw=1, ms=5)
ax[0].plot(t, rho, color="C0", lw=1)
ax[0].set_ylabel(r'$\rho\,$ ["]')
ax[0].set_title("initial orbit")

ax[1].errorbar(astro_jds, theta_data, yerr=theta_err, fmt=".k", lw=1, ms=5)
ax[1].plot(t, theta, color="C0", lw=1)
ax[1].set_ylabel(r"P.A. [radians]")
_ = ax[1].set_xlabel("time [JD]")
# -

# Now that we have an initial orbit, we can set the model up using PyMC3 to do inference.

# +
yr = 365.25

# for predicted orbits
t_fine = np.linspace(astro_jds.min() - 500, astro_jds.max() + 500, num=1000)


def get_model(parallax=None):
    with pm.Model() as model:

        if parallax is None:
            # Without an actual parallax measurement, we can model the orbit in units of arcseconds
            # by providing a fake_parallax and conversion constant
            plx = 1  # arcsec
        else:
            # Below we will run a version of this model where a measurement of parallax is provided
            # The measurement is in milliarcsec
            m_plx = pm.Bound(pm.Normal, lower=0, upper=100)(
                "m_plx", mu=parallax[0], sd=parallax[1], testval=parallax[0]
            )
            plx = pm.Deterministic("plx", 1e-3 * m_plx)

        a_ang = pm.Uniform("a_ang", 0.1, 1.0, testval=0.324)
        a = pm.Deterministic("a", a_ang / plx)

        # We expect the period to be somewhere in the range of 25 years,
        # so we'll set a broad prior on logP
        logP = pm.Normal(
            "logP", mu=np.log(25 * yr), sd=10.0, testval=np.log(28.8 * yr)
        )
        P = pm.Deterministic("P", tt.exp(logP))

        # For astrometric-only fits, it's generally better to fit in
        # p = (Omega + omega)/2 and m = (Omega - omega)/2 instead of omega and Omega
        # directly
        omega0 = 251.6 * deg - np.pi
        Omega0 = 159.6 * deg
        p = Angle("p", testval=0.5 * (Omega0 + omega0))
        m = Angle("m", testval=0.5 * (Omega0 - omega0))
        omega = pm.Deterministic("omega", p - m)
        Omega = pm.Deterministic("Omega", p + m)

        # For these orbits, it can also be better to fit for a phase angle
        # (relative to a reference time) instead of the time of periasteron
        # passage directly
        phase = Angle("phase", testval=0.0)
        tperi = pm.Deterministic("tperi", T0 + P * phase / (2 * np.pi))

        # Geometric uiform prior on cos(incl)
        cos_incl = pm.Uniform(
            "cos_incl", lower=-1, upper=1, testval=np.cos(96.0 * deg)
        )
        incl = pm.Deterministic("incl", tt.arccos(cos_incl))
        ecc = pm.Uniform("ecc", lower=0.0, upper=1.0, testval=0.798)

        # Set up the orbit
        orbit = xo.orbits.KeplerianOrbit(
            a=a * au_to_R_sun,
            t_periastron=tperi,
            period=P,
            incl=incl,
            ecc=ecc,
            omega=omega,
            Omega=Omega,
        )
        if parallax is not None:
            pm.Deterministic("M_tot", orbit.m_total)

        # Compute the model in rho and theta
        rho_model, theta_model = orbit.get_relative_angles(astro_jds, plx)
        pm.Deterministic("rho_model", rho_model)
        pm.Deterministic("theta_model", theta_model)

        # Add jitter terms to both separation and position angle
        log_rho_s = pm.Normal(
            "log_rho_s", mu=np.log(np.median(rho_err)), sd=5.0
        )
        log_theta_s = pm.Normal(
            "log_theta_s", mu=np.log(np.median(theta_err)), sd=5.0
        )
        rho_tot_err = tt.sqrt(rho_err ** 2 + tt.exp(2 * log_rho_s))
        theta_tot_err = tt.sqrt(theta_err ** 2 + tt.exp(2 * log_theta_s))

        # define the likelihood function, e.g., a Gaussian on both rho and theta
        pm.Normal("rho_obs", mu=rho_model, sd=rho_tot_err, observed=rho_data)

        # We want to be cognizant of the fact that theta wraps so the following is equivalent to
        # pm.Normal("obs_theta", mu=theta_model, observed=theta_data, sd=theta_tot_err)
        # but takes into account the wrapping. Thanks to Rob de Rosa for the tip.
        theta_diff = tt.arctan2(
            tt.sin(theta_model - theta_data), tt.cos(theta_model - theta_data)
        )
        pm.Normal("theta_obs", mu=theta_diff, sd=theta_tot_err, observed=0.0)

        # Set up predicted orbits for later plotting
        rho_dense, theta_dense = orbit.get_relative_angles(t_fine, plx)
        rho_save = pm.Deterministic("rho_save", rho_dense)
        theta_save = pm.Deterministic("theta_save", theta_dense)

        # Optimize to find the initial parameters
        map_soln = model.test_point
        map_soln = xo.optimize(map_soln, vars=[log_rho_s, log_theta_s])
        map_soln = xo.optimize(map_soln, vars=[phase])
        map_soln = xo.optimize(map_soln, vars=[p, m, ecc])
        map_soln = xo.optimize(map_soln, vars=[logP, a_ang, phase])
        map_soln = xo.optimize(map_soln)

    return model, map_soln


model, map_soln = get_model()
# -

# Now that we have a maximum a posteriori estimate of the parameters, let's take a look at the results to make sure that they seem reasonable.

# +
ekw = dict(fmt=".k", lw=0.5)

fig, ax = plt.subplots(nrows=4, sharex=True, figsize=(6, 8))
ax[0].set_ylabel(r'$\rho\,$ ["]')
ax[1].set_ylabel(r"$\rho$ residuals")
ax[2].set_ylabel(r"P.A. [radians]")
ax[3].set_ylabel(r"P.A. residuals")

tot_rho_err = np.sqrt(rho_err ** 2 + np.exp(2 * map_soln["log_rho_s"]))
tot_theta_err = np.sqrt(theta_err ** 2 + np.exp(2 * map_soln["log_theta_s"]))

ax[0].errorbar(astro_jds, rho_data, yerr=tot_rho_err, **ekw)
ax[0].plot(t_fine, map_soln["rho_save"], "C1")

ax[1].axhline(0.0, color="0.5")
ax[1].errorbar(
    astro_jds, rho_data - map_soln["rho_model"], yerr=tot_rho_err, **ekw
)


ax[2].plot(t_fine, map_soln["theta_save"], "C1")
ax[2].errorbar(astro_jds, theta_data, yerr=tot_theta_err, **ekw)

ax[3].axhline(0.0, color="0.5")
ax[3].errorbar(
    astro_jds, theta_data - map_soln["theta_model"], yerr=tot_theta_err, **ekw
)

ax[3].set_xlim(t_fine[0], t_fine[-1])
_ = ax[0].set_title("map orbit")
# -

# Now let's sample the posterior.

np.random.seed(1234)
with model:
    trace = pm.sample(
        tune=5000,
        draws=4000,
        start=map_soln,
        cores=2,
        chains=2,
        init="adapt_full",
        target_accept=0.9,
    )

# First we can check the convergence for some of the key parameters.

with model:
    summary = pm.summary(
        trace,
        varnames=["P", "tperi", "a_ang", "omega", "Omega", "incl", "ecc"],
    )
summary

# That looks pretty good.
# Now here's a corner plot showing the covariances between parameters.

# +
import corner

samples = pm.trace_to_dataframe(trace, varnames=["ecc"])
samples["$P$ [yr]"] = trace["P"] / yr
samples["$T_\mathrm{peri} - T_0$ [day]"] = trace["tperi"] - T0
samples["$a$ [arcsec]"] = trace["a_ang"]
samples["$\omega$ [deg]"] = (trace["omega"] / deg) % 360
samples["$\Omega$ [deg]"] = (trace["Omega"] / deg) % 360
samples["$i$ [deg]"] = (trace["incl"] / deg) % 360
samples["$e$"] = samples["ecc"]
del samples["ecc"]
_ = corner.corner(samples)
# -

# Finally, we can plot the posterior constraints on $\rho$ and $\theta$.
# This figure is much like the one for the MAP solution above, but this time the orange is a contour (not a line) showing the 68% credible region for the model.

# +
ekw = dict(fmt=".k", lw=0.5)

fig, ax = plt.subplots(nrows=2, sharex=True, figsize=(6, 6))
ax[0].set_ylabel(r'$\rho\,$ ["]')
ax[1].set_ylabel(r"P.A. [radians]")

tot_rho_err = np.sqrt(
    rho_err ** 2 + np.exp(2 * np.median(trace["log_rho_s"], axis=0))
)
tot_theta_err = np.sqrt(
    theta_err ** 2 + np.exp(2 * np.median(trace["log_theta_s"], axis=0))
)

ax[0].errorbar(astro_jds, rho_data, yerr=tot_rho_err, **ekw)
q = np.percentile(trace["rho_save"], [16, 84], axis=0)
ax[0].fill_between(t_fine, q[0], q[1], color="C1", alpha=0.8, lw=0)

ax[1].errorbar(astro_jds, theta_data, yerr=tot_theta_err, **ekw)
q = np.percentile(trace["theta_save"], [16, 84], axis=0)
ax[1].fill_between(t_fine, q[0], q[1], color="C1", alpha=0.8, lw=0)

ax[-1].set_xlim(t_fine[0], t_fine[-1])
_ = ax[0].set_title("posterior inferences")
# -

# As we can see from the narrow range of orbits (the orange swath appears like a thin line), the orbit is actually highly constrained by the astrometry.
# We also see two outlier epochs in the vicinity of 2445000 - 2447000, since adjacent epochs seem to be right on the orbit.
# It's likely the uncertainties were not estimated correctly for these, and the simplistic jitter model we implemented isn't sophisticated to apply more weight to only these discrepant points.

# ## Including parallax
#
# While this is encouraging that we fit an astrometric orbit, a simple astrometric fit to just $\rho$ and $\theta$ isn't actually that physically satisfying, since many of the orbital parameters simply have to do with the orientation relative to us ($i$, $\omega$, and $\Omega$). The only truly intrinsic parameters are $P$ and $e$. To learn more about some of the physical parameters, such as the total mass of the system, we'd like to incorporate distance information to put a physical scale to the problem.
#
# The *Gaia* DR2 parallax is $\varpi = 24.05 \pm 0.45$ mas.
#
# We can use exactly the same model as above with only an added parallax constraint:

plx_model, plx_map_soln = get_model(parallax=[24.05, 0.45])

np.random.seed(5432)
with plx_model:
    plx_trace = pm.sample(
        tune=5000,
        draws=4000,
        start=plx_map_soln,
        cores=2,
        chains=2,
        init="adapt_full",
        target_accept=0.9,
    )

# Check the convergence diagnostics.

with model:
    summary = pm.summary(
        plx_trace,
        varnames=[
            "P",
            "tperi",
            "a_ang",
            "omega",
            "Omega",
            "incl",
            "ecc",
            "M_tot",
            "plx",
        ],
    )
summary

# And make the corner plot for the physical parameters.

samples = pm.trace_to_dataframe(plx_trace, varnames=["ecc"])
samples["$P$ [yr]"] = plx_trace["P"] / yr
samples["$T_\mathrm{peri} - T_0$ [day]"] = plx_trace["tperi"] - T0
samples["$a$ [au]"] = plx_trace["a"]
samples["$M_\mathrm{tot}$ [$M_\odot$]"] = plx_trace["M_tot"]
samples["$e$"] = plx_trace["ecc"]
del samples["ecc"]
_ = corner.corner(samples)

# ## Citations
#
# As described in the [Citing exoplanet & its dependencies](./citation.ipynb) tutorial, we can use :func:`exoplanet.citations.get_citations_for_model` to construct an acknowledgment and BibTeX listing that includes the relevant citations for this model.

with model:
    txt, bib = xo.citations.get_citations_for_model()
print(txt)

print("\n".join(bib.splitlines()[:10]) + "\n...")



# -*- coding: utf-8 -*-

import numpy as np
import pytest
import theano

from exoplanet.light_curves import LimbDarkLightCurve
from exoplanet.orbits.keplerian import KeplerianOrbit
from exoplanet.orbits.rebound import ReboundOrbit

try:
    import rebound  # NOQA
except ImportError:
    pytest.skip(
        "skipping rebound tests when not installed", allow_module_level=True
    )


@pytest.mark.parametrize(
    "orbit",
    (
        ReboundOrbit(
            period=50.0,
            t0=1.55,
            ecc=0.28,
            omega=-4.56,
            m_planet=0.12,
            b=0.51,
            m_star=1.51,
            r_star=1.0,
        ),
        ReboundOrbit(
            period=[50.0, 87.5],
            t0=[1.55, 10.6],
            ecc=[0.28, 0.01],
            omega=[-4.56, 1.5],
            m_planet=[0.0, 0.0],
            b=[0.51, 0.21],
            m_star=1.51,
            r_star=1.0,
        ),
    ),
)
def test_keplerian(orbit):
    t = np.linspace(50, 1000, 1045)

    x, y, z = theano.function([], orbit.get_relative_position(t))()
    x0, y0, z0 = theano.function(
        [], super(ReboundOrbit, orbit).get_relative_position(t)
    )()
    assert np.allclose(x, x0)
    assert np.allclose(y, y0)
    assert np.allclose(z, z0)

    x, y, z = theano.function([], orbit.get_planet_position(t))()
    x0, y0, z0 = theano.function(
        [], super(ReboundOrbit, orbit).get_planet_position(t)
    )()
    assert np.allclose(x, x0)
    assert np.allclose(y, y0)
    assert np.allclose(z, z0)

    x, y, z = theano.function([], orbit.get_star_position(t))()
    x0, y0, z0 = theano.function(
        [], super(ReboundOrbit, orbit).get_star_position(t)
    )()
    if len(np.shape(x0)) > 1:
        assert np.allclose(x, np.sum(x0, axis=-1))
        assert np.allclose(y, np.sum(y0, axis=-1))
        assert np.allclose(z, np.sum(z0, axis=-1))
    else:
        assert np.allclose(x, x0)
        assert np.allclose(y, y0)
        assert np.allclose(z, z0)


@pytest.mark.xfail(reason="I don't understand Theano sometimes")
def test_tensor_bug():
    orbit = ReboundOrbit(
        period=50.0, t0=0.0, ecc=0.5, omega=0.1, b=0.2, m_planet=0.5
    )
    orbit2 = ReboundOrbit(
        period=50.0,
        t0=0.0,
        ecc=0.5 + 1e-9,
        omega=0.1,
        b=0.2,
        m_planet=0.5 - 1e-9,
    )
    t = np.linspace(50, 1000, 1045)

    x, y, z = theano.function([], orbit.get_relative_position(t))()
    x2, y2, z2 = theano.function([], orbit2.get_relative_position(t))()

    x0, y0, z0 = theano.function(
        [], super(ReboundOrbit, orbit).get_relative_position(t)
    )()
    x02, y02, z02 = theano.function(
        [], super(ReboundOrbit, orbit2).get_relative_position(t)
    )()

    assert np.allclose(x0, x02)
    assert np.allclose(y0, y02)
    assert np.allclose(z0, z02)

    assert np.allclose(x, x2)
    assert np.allclose(y, y2)
    assert np.allclose(z, z2)


def test_keplerian_light_curve():
    t = np.linspace(50, 1000, 1045)
    r = np.array([0.04, 0.02])
    args = dict(
        period=[50.0, 87.5],
        t0=[1.55, 10.6],
        ecc=[0.28, 0.01],
        omega=[-4.56, 1.5],
        m_planet=[0.0, 0.0],
        b=[0.51, 0.21],
        m_star=1.51,
        r_star=1.0,
    )
    orbit0 = KeplerianOrbit(**args)
    orbit = ReboundOrbit(**args)

    ld = LimbDarkLightCurve([0.2, 0.3])
    lc0 = ld.get_light_curve(orbit=orbit0, r=r, t=t).eval()
    lc = ld.get_light_curve(orbit=orbit, r=r, t=t).eval()

    assert np.allclose(lc0, lc)



# -*- coding: utf-8 -*-

__all__ = ["ReboundOrbit"]

import theano.tensor as tt

from .constants import au_per_R_sun, day_per_yr_over_2pi
from .keplerian import KeplerianOrbit

try:
    from rebound_pymc3.integrate import IntegrateOp as ReboundOp
except ImportError:
    from ..theano_ops.rebound import ReboundOp


class ReboundOrbit(KeplerianOrbit):
    """An N-body system powered by the rebound integrator

    This takes all the same arguments as the :class:`KeplerianOrbit`, but
    these arguments define the orbital elements at some reference time (given
    by the ``rebound_t`` parameter). The positions and velocities of the bodies
    are then computed by numerically integrating the gravitational N-body
    system.

    ``rebound``-specific parameters can be provided as keyword arguments
    prefixed by ``rebound_``. These will then be applied to the
    ``rebound.Simulation`` object as properties. Therefore, if you want to
    change the integrator, you could use: ``rebound_integrator = "whfast"``,
    for example. All of these parameters are passed directly through to
    ``rebound`` except ``rebound_t`` (the reference time) which is converted
    from days to years over two pi (the default time units in ``rebound``).

    .. note:: ``exoplanet`` and ``rebound`` use different base units, but all
        of the unit conversions are handled behind the scenes in this object
        so that means that you should mostly use the ``exoplanet`` units when
        interacting with this class and you should be very cautious about
        setting the ``rebound_G`` argument. One example of a case where you'll
        need to use the ``rebound`` units is when you want to set the
        integrator step size using the ``rebound_dt`` parameter.

    """

    __citations__ = ("astropy", "rebound")

    def __init__(self, *args, **kwargs):
        rebound_args = dict(
            (k[8:], kwargs.pop(k))
            for k in list(kwargs.keys())
            if k.startswith("rebound_")
        )
        # rebound_args["G"] = rebound_G  # Consistent units
        rebound_args["t"] = rebound_args.get("t", 0.0) / day_per_yr_over_2pi
        self.rebound_initial_time = rebound_args["t"]
        self.rebound_op = ReboundOp(**rebound_args)
        super(ReboundOrbit, self).__init__(*args, **kwargs)

        if self.period.ndim:
            self.masses = tt.concatenate(
                [[self.m_star], self.m_planet + tt.zeros_like(self.period)]
            )
        else:
            self.masses = tt.stack([self.m_star, self.m_planet])

    def _get_position_and_velocity(self, t):
        pos, vel = super(ReboundOrbit, self)._get_position_and_velocity(
            self.rebound_initial_time
        )
        initial_coords = tt.concatenate(
            (pos * au_per_R_sun, vel * au_per_R_sun * day_per_yr_over_2pi),
            axis=-1,
        )
        coords, _ = self.rebound_op(
            self.masses,
            initial_coords,
            (t - self.rebound_initial_time) / day_per_yr_over_2pi,
        )

        # Deal with strange ordering and the need for things to be C contiguous
        coords = coords.dimshuffle(2, 0, 1)
        coords = coords + tt.zeros(coords.shape, dtype=coords.dtype)

        pos = coords[:3, :, :] / au_per_R_sun
        vel = coords[3:, :, :] / (day_per_yr_over_2pi * au_per_R_sun)
        return pos, vel

    def get_planet_position(self, t, light_delay=False):
        """The planets' positions in the barycentric frame

        Args:
            t: The times where the position should be evaluated.

        Returns:
            The components of the position vector at ``t`` in units of
            ``R_sun``.

        """
        if light_delay:
            raise NotImplementedError(
                "Light travel time delay is not implemented for rebound orbits"
            )
        pos, _ = self._get_position_and_velocity(t)
        if self.period.ndim:
            return pos[0, :, 1:], pos[1, :, 1:], pos[2, :, 1:]
        return pos[0, :, 1], pos[1, :, 1], pos[2, :, 1]

    def get_star_position(self, t, light_delay=False):
        """The star's position in the barycentric frame

        .. note:: Unlike the :class:`KeplerianOrbit`, this will not return
            the contributions from each planet separately.

        Args:
            t: The times where the position should be evaluated.

        Returns:
            The components of the position vector at ``t`` in units of
            ``R_sun``.

        """
        if light_delay:
            raise NotImplementedError(
                "Light travel time delay is not implemented for rebound orbits"
            )
        pos, _ = self._get_position_and_velocity(t)
        return pos[0, :, 0], pos[1, :, 0], pos[2, :, 0]

    def get_relative_position(self, t, light_delay=False):
        """The planets' positions relative to the star in the X,Y,Z frame.

        Args:
            t: The times where the position should be evaluated.

        Returns:
            The components of the position vector at ``t`` in units of
            ``R_sun``.

        """
        if light_delay:
            raise NotImplementedError(
                "Light travel time delay is not implemented for rebound orbits"
            )
        pos, _ = self._get_position_and_velocity(t)
        if self.period.ndim:
            pos = pos[:, :, 1:] - pos[:, :, 0][:, :, None]
        else:
            pos = pos[:, :, 1] - pos[:, :, 0]
        return pos[0], pos[1], pos[2]

    def get_planet_velocity(self, t):
        """Get the planets' velocity vectors

        Args:
            t: The times where the velocity should be evaluated.

        Returns:
            The components of the velocity vector at ``t`` in units of
            ``M_sun/day``.

        """
        _, vel = self._get_position_and_velocity(t)
        if self.period.ndim:
            return vel[0, :, 1:], vel[1, :, 1:], vel[2, :, 1:]
        return vel[0, :, 1], vel[1, :, 1], vel[2, :, 1]

    def get_star_velocity(self, t):
        """Get the star's velocity vector

        .. note:: Unlike the :class:`KeplerianOrbit`, this will not return
            the contributions from each planet separately.

        Args:
            t: The times where the velocity should be evaluated.

        Returns:
            The components of the velocity vector at ``t`` in units of
            ``M_sun/day``.

        """
        _, vel = self._get_position_and_velocity(t)
        return vel[0, :, 0], vel[1, :, 0], vel[2, :, 0]

    def get_relative_velocity(self, t):
        """The planets' velocity relative to the star

        Args:
            t: The times where the velocity should be evaluated.

        Returns:
            The components of the velocity vector at ``t`` in units of
            ``R_sun/day``.

        """
        _, vel = self._get_position_and_velocity(t)
        if self.period.ndim:
            vel = vel[:, :, 1:] - vel[:, :, 0][:, :, None]
        else:
            vel = vel[:, :, 1] - vel[:, :, 0]
        return vel[0], vel[1], vel[2]

    def _get_acceleration(self, *args, **kwargs):
        raise NotImplementedError(
            "the ReboundOrbit doesn't compute accelerations"
        )

    def get_radial_velocity(self, t, output_units=None):
        """Get the radial velocity of the star

        .. note:: The convention in exoplanet is that positive `z` points
            *towards* the observer. However, for consistency with radial
            velocity literature this method returns values where positive
            radial velocity corresponds to a redshift as expected.

        .. note:: Unlike the :class:`KeplerianOrbit` implementation, the
            semi-amplitude ``K`` cannot be used with the :class:`ReboundOrbit`.
            Also, the contributions of each planet are not returned separately;
            this will always return a single time series.

        Args:
            t: The times where the radial velocity should be evaluated.
            output_units (Optional): An AstroPy velocity unit. If not given,
                the output will be evaluated in ``m/s``. This is ignored if a
                value is given for ``K``.

        Returns:
            The reflex radial velocity evaluated at ``t`` in units of
            ``output_units``.

        """
        return super(ReboundOrbit, self).get_radial_velocity(
            t, K=None, output_units=output_units
        )

    def in_transit(self, t, r=0.0, texp=None, light_delay=False):
        """This is a no-op and all points are assumed to be in transit"""
        return tt.arange(t.size)



# -*- coding: utf-8 -*-

__all__ = ["RadiusImpact", "get_joint_radius_impact"]

import warnings

import numpy as np
import pymc3 as pm
import theano.tensor as tt
from pymc3.distributions import draw_values, generate_samples

from ..citations import add_citations_to_model
from ..utils import deprecated
from . import transforms as tr


class RadiusImpact(pm.Flat):
    """The Espinoza (2018) distribution over radius and impact parameter

    This is an implementation of `Espinoza (2018)
    <http://iopscience.iop.org/article/10.3847/2515-5172/aaef38/meta>`_
    The first axis of the shape of the parameter should be exactly 2. The
    radius ratio will be in the zeroth entry in the first dimension and
    the impact parameter will be in the first.

    Args:
        min_radius: The minimum allowed radius.
        max_radius: The maximum allowed radius.

    """

    __citations__ = ("espinoza18",)

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "'RadiusImpact' is deprecated. Use 'ImpactParameter' instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        add_citations_to_model(self.__citations__, kwargs.get("model", None))

        # Make sure that the shape is compatible
        shape = kwargs.get("shape", 2)
        try:
            if list(shape)[0] != 2:
                raise ValueError("the first dimension should be exactly 2")
        except TypeError:
            if shape != 2:
                raise ValueError("the first dimension should be exactly 2")

        self.min_radius = kwargs.pop("min_radius", 0)
        self.max_radius = kwargs.pop("max_radius", 1)
        transform = tr.radius_impact(self.min_radius, self.max_radius)
        kwargs["shape"] = shape
        kwargs["transform"] = transform

        super(RadiusImpact, self).__init__(*args, **kwargs)

        # Work out some reasonable starting values for the parameters
        default = np.zeros(shape)
        mn, mx = draw_values([self.min_radius - 0.0, self.max_radius - 0.0])
        default[0] = 0.5 * (mn + mx)
        default[1] = 0.5
        self._default = default

    def _random(self, pl, pu, size=None):
        r = np.moveaxis(
            np.random.uniform(0, 1, size=size), 0, -len(self.shape)
        )

        dr = pu - pl
        denom = 2 + pu + pl
        Ar = dr / denom

        r1 = r[0]
        r2 = r[1]
        m = r1 > Ar

        b = np.empty_like(r1)
        p = np.empty_like(r1)

        b[m] = (1 + pl) * (1 + (r1[m] - 1) / (1 - Ar))
        p[m] = pl + r2[m] * dr

        q1 = r1[~m] / Ar
        q2 = r2[~m]
        b[~m] = (1 + pl) + np.sqrt(q1) * q2 * dr
        p[~m] = pu - dr * np.sqrt(q1) * (1 - q2)

        pb = np.stack([p, b], axis=0)

        return np.moveaxis(pb, 0, -len(self.shape))

    def random(self, point=None, size=None):
        mn, mx = draw_values(
            [self.min_radius - 0.0, self.max_radius - 0.0], point=point
        )
        return generate_samples(
            self._random,
            mn,
            mx,
            dist_shape=self.shape,
            broadcast_shape=self.shape,
            size=size,
        )

    def logp(self, value):
        return tt.zeros_like(tt.as_tensor_variable(value))


@deprecated("ImpactParameter")
def get_joint_radius_impact(
    name="",
    N_planets=None,
    min_radius=0,
    max_radius=1,
    testval_r=None,
    testval_b=None,
    **kwargs,
):
    """Get the joint distribution over radius and impact parameter

    This uses the Espinoza (2018) parameterization of the distribution (see
    :class:`distributions.RadiusImpact` for more details).

    Args:
        name (Optional[str]): A prefix that is added to all distribution names
            used in this parameterization. For example, if ``name`` is
            ``param_``, vars will be added to the PyMC3 model with names
            ``param_rb`` (for the joint distribution), ``param_b``, and
            ``param_r``.
        N_planets (Optional[int]): The number of planets. If not provided, it
            will be inferred from the ``testval_*`` parameters or assumed to
            be 1.
        min_radius (Optional[float]): The minimum allowed radius.
        max_radius (Optional[float]): The maximum allowed radius.
        testval_r (Optional[float or array]): An initial guess for the radius
            parameter. This should be a ``float`` or an array with
            ``N_planets`` entries.
        testval_b (Optional[float or array]): An initial guess for the impact
            parameter. This should be a ``float`` or an array with
            ``N_planets`` entries.

    Returns:
        Two ``pymc3.Deterministic`` variables for the planet radius and impact
        parameter.

    """
    if N_planets is None:
        if testval_r is not None:
            N_planets = len(np.atleast_1d(testval_r))
        elif testval_b is not None:
            N_planets = len(np.atleast_1d(testval_b))
        else:
            N_planets = 1
    N_planets = int(N_planets)

    # Set up the testval for the rb parameter
    rb_test = np.zeros((2, N_planets))
    if testval_r is None:
        rb_test[0, :] = 0.5 * (min_radius + max_radius)
    else:
        rb_test[0, :] = testval_r
    if testval_b is None:
        rb_test[1, :] = 0.5
    else:
        rb_test[1, :] = testval_b

    # Construct the join distribution
    rb = RadiusImpact(
        name + "rb",
        min_radius=min_radius,
        max_radius=max_radius,
        shape=(2, N_planets),
        testval=rb_test,
        **kwargs,
    )

    # Extract the individual components
    r = pm.Deterministic(name + "r", rb[0])
    b = pm.Deterministic(name + "b", rb[1])

    return r, b



# -*- coding: utf-8 -*-

import astropy.units as u
import numpy as np

from exoplanet.estimators import (
    autocorr_estimator,
    bls_estimator,
    estimate_minimum_mass,
    estimate_semi_amplitude,
    lomb_scargle_estimator,
)
from exoplanet.orbits import KeplerianOrbit


def test_estimate_semi_amplitude(seed=9502):
    np.random.seed(seed)
    t = np.sort(np.random.uniform(0, 10, 500))
    period = 2.345
    amp = 4.5
    y = amp * np.sin(2 * np.pi * t / period)
    est = estimate_semi_amplitude(period, t, y)
    assert np.allclose(est, amp)

    est = estimate_semi_amplitude(
        period, t, y, yerr=np.ones_like(t), t0s=0.5 * period
    )
    assert np.allclose(est, amp)


def test_estimate_minimum_mass(seed=9502):
    np.random.seed(seed)
    t = np.sort(np.random.uniform(0, 10, 500))
    period = 2.345
    t0 = 0.5
    orbit = KeplerianOrbit(period=period, t0=t0, m_planet=0.01, incl=0.8)
    y = orbit.get_radial_velocity(t).eval()
    m1 = (orbit.m_planet * orbit.sin_incl).eval()
    m2 = estimate_minimum_mass(period, t, y).to(u.M_sun).value
    m3 = estimate_minimum_mass(period, t, y, t0s=t0).to(u.M_sun).value
    assert np.abs((m1 - m2) / m1) < 0.01
    assert np.abs((m1 - m3) / m1) < 0.01


def test_lomb_scargle_estimator(seed=9502):
    np.random.seed(seed)
    t = np.sort(np.random.uniform(0, 10, 500))
    period = 2.345
    amp = 4.5
    y = amp * np.sin(2 * np.pi * t / period)
    results = lomb_scargle_estimator(
        t, y, min_period=1, max_period=10, filter_period=10.0
    )
    assert np.abs(period - results["peaks"][0]["period"]) / period < 0.001


def test_autocorr_estimator(seed=9502):
    np.random.seed(seed)
    t = np.linspace(0, 10, 500)
    period = 2.345
    amp = 4.5
    y = amp * np.sin(2 * np.pi * t / period)
    results = autocorr_estimator(
        t, y, min_period=0.01, max_period=10, smooth=0.0
    )
    assert np.abs(period - results["peaks"][0]["period"]) / period < 0.01


def test_bls_estimator(seed=721):
    np.random.seed(seed)
    t = np.sort(np.random.uniform(0, 27, 5000))
    y = 1e-3 * np.random.randn(len(t))
    period = 2.345
    t0 = 1.0
    depth = 0.1
    duration = 0.2
    y[
        np.abs((t - t0 + 0.5 * period) % period - 0.5 * period)
        < 0.5 * duration
    ] -= depth
    results = bls_estimator(t, y)
    assert np.abs(period - results["peaks"][0]["period"]) / period < 0.01



# -*- coding: utf-8 -*-

__all__ = ["with_unit", "has_unit", "to_unit"]

from .utils import as_tensor_variable

UNIT_ATTR_NAME = "__exoplanet_unit__"


def with_unit(obj, unit):
    """Decorate a Theano tensor with Astropy units

    Args:
        obj: The Theano tensor
        unit (astropy.Unit): The units for this object

    Raises:
        TypeError: If the tensor already has units

    """
    if hasattr(obj, UNIT_ATTR_NAME):
        raise TypeError("{0} already has units".format(repr(obj)))
    obj = as_tensor_variable(obj)
    setattr(obj, UNIT_ATTR_NAME, unit)
    return obj


def has_unit(obj):
    """Does an object have units as defined by exoplanet?"""
    return hasattr(obj, UNIT_ATTR_NAME)


def to_unit(obj, target):
    """Convert a Theano tensor with units to a target set of units

    Args:
        obj: The Theano tensor
        target (astropy.Unit): The target units

    Returns:
        A Theano tensor in the right units

    """
    if not has_unit(obj):
        return obj
    base = getattr(obj, UNIT_ATTR_NAME)
    return obj * base.to(target)



import numpy as np
import pymc3 as pm

from exoplanet.utils import eval_in_model


def test_eval_in_model(seed=123409):
    np.random.seed(seed)
    x_val = np.random.randn(5, 3)
    x_val2 = np.random.randn(5, 3)
    with pm.Model():
        x = pm.Normal("x", shape=x_val.shape, testval=x_val)
        assert np.allclose(eval_in_model(x), x_val)
        assert np.allclose(eval_in_model(x, {"x": x_val2}), x_val2)



# -*- coding: utf-8 -*-

__all__ = [
    "logger",
    "eval_in_model",
    "get_samples_from_trace",
    "get_args_for_theano_function",
    "get_theano_function_for_var",
    "deprecation_warning",
    "deprecated",
]

import logging
import warnings
from functools import wraps

import numpy as np
import pymc3 as pm
import theano

logger = logging.getLogger("exoplanet")

if theano.config.floatX != "float64":
    warnings.warn(
        "exoplanet should only be used with 'float64' precision, "
        "but theano.config.floatX == '{0}'".format(theano.config.floatX)
    )


def as_tensor_variable(x, dtype="float64", **kwargs):
    t = theano.tensor.as_tensor_variable(x, **kwargs)
    if dtype is None:
        return t
    return t.astype(dtype)


def get_args_for_theano_function(point=None, model=None):
    """Get the arguments required to evaluate a PyMC3 model component

    Use the result as the arguments for the callable returned by the
    :func:`get_theano_function_for_var` function.

    Args:
        point (dict, optional): The point in parameter space where the model
            componenet should be evaluated
        model (optional): The PyMC3 model object. By default, the current
            context will be used.

    """
    model = pm.modelcontext(model)
    if point is None:
        point = model.test_point
    return [point[k.name] for k in model.vars]


def get_theano_function_for_var(var, model=None, **kwargs):
    """Get a callable function to evaluate a component of a PyMC3 model

    This should then be called using the arguments returned by the
    :func:`get_args_for_theano_function` function.

    Args:
        var: The model component to evaluate. This can be a Theano tensor, a
            PyMC3 variable, or a list of these
        model (optional): The PyMC3 model object. By default, the current
            context will be used.

    """
    model = pm.modelcontext(model)
    kwargs["on_unused_input"] = kwargs.get("on_unused_input", "ignore")
    return theano.function(model.vars, var, **kwargs)


def eval_in_model(var, point=None, return_func=False, model=None, **kwargs):
    """Evaluate a Theano tensor or PyMC3 variable in a PyMC3 model

    This method builds a Theano function for evaluating a node in the graph
    given the required parameters. This will also cache the compiled Theano
    function in the current ``pymc3.Model`` to reduce the overhead of calling
    this function many times.

    Args:
        var: The variable or tensor to evaluate.
        point (Optional): A ``dict`` of input parameter values. This can be
            ``model.test_point`` (default), the result of ``pymc3.find_MAP``,
            a point in a ``pymc3.MultiTrace`` or any other representation of
            the input parameters.
        return_func (Optional[bool]): If ``False`` (default), return the
            evaluated variable. If ``True``, return the result, the Theano
            function and the list of arguments for that function.

    Returns:
        Depending on ``return_func``, either the value of ``var`` at ``point``,
        or this value, the Theano function, and the input arguments.

    """
    func = get_theano_function_for_var(var, model=model, **kwargs)
    args = get_args_for_theano_function(point=point, model=model)
    if return_func:
        return func(*args), func, args
    return func(*args)


def get_samples_from_trace(trace, size=1):
    """Generate random samples from a PyMC3 MultiTrace

    Args:
        trace: The ``MultiTrace``.
        size: The number of samples to generate.

    """
    for i in range(size):
        chain_idx = np.random.randint(len(trace.chains))
        sample_idx = np.random.randint(len(trace))
        yield trace._straces[chain_idx][sample_idx]


def deprecation_warning(msg):
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)


def deprecated(alternate=None):
    def wrapper(func, alternate=alternate):
        msg = "'{0}' is deprecated.".format(func.__name__)
        if alternate is not None:
            msg += " Use '{0}' instead.".format(alternate)

        @wraps(func)
        def f(*args, **kwargs):
            deprecation_warning(msg)
            return func(*args, **kwargs)

        return f

    return wrapper



# -*- coding: utf-8 -*-

import pickle

import numpy as np
import theano
import theano.tensor as tt

from exoplanet.theano_ops.driver import SimpleLimbDark
from exoplanet.theano_ops.starry import (
    GetCl,
    GetClRev,
    LimbDark,
    RadiusFromOccArea,
)
from exoplanet.theano_ops.test_tools import InferShapeTester


class TestGetCl(InferShapeTester):
    def setup_method(self):
        super().setup_method()
        self.op_class = GetCl
        self.op = GetCl()

    def test_basic(self):
        x = tt.dvector()
        f = theano.function([x], self.op(x))

        inp = np.array([-1, 0.3, 0.2, 0.5])
        out = f(inp)

        assert np.allclose(np.array([-0.85, 2.5, -0.425, 0.1]), out)

    def test_infer_shape(self):
        x = tt.dvector()
        self._compile_and_check(
            [x], [self.op(x)], [np.asarray(np.random.rand(5))], self.op_class
        )

    def test_grad(self):
        tt.verify_grad(self.op, [np.array([-1, 0.3, 0.2, 0.5])], rng=np.random)


class TestGetClRev(InferShapeTester):
    def setup_method(self):
        super().setup_method()
        self.op_class = GetClRev
        self.op = GetClRev()

    def test_basic(self):
        x = tt.dvector()
        f = theano.function([x], self.op(x))

        inp = np.array([-1, 0.3, 0.2, 0.5])
        out = f(inp)

        assert np.allclose(np.array([0, 1.3, 2.05, 3.53]), out)

    def test_infer_shape(self):
        x = tt.dvector()
        self._compile_and_check(
            [x], [self.op(x)], [np.asarray(np.random.rand(5))], self.op_class
        )


class TestLimbDark(InferShapeTester):
    def setup_method(self):
        super().setup_method()
        self.op_class = LimbDark
        self.op = LimbDark()

    def get_args(self):
        c = tt.vector()
        b = tt.vector()
        r = tt.vector()
        los = tt.vector()
        f = theano.function([c, b, r, los], self.op(c, b, r, los)[0])

        c_val = np.array([-0.85, 2.5, -0.425, 0.1])
        b_val = np.linspace(-1.5, 1.5, 100)
        r_val = 0.1 + np.zeros_like(b_val)
        los_val = np.ones_like(b_val)

        return f, [c, b, r, los], [c_val, b_val, r_val, los_val]

    def test_basic(self):
        f, _, in_args = self.get_args()
        out = f(*in_args)
        assert np.allclose(0.0, out[0])
        assert np.allclose(0.0, out[-1])

    def test_los(self):
        f, _, in_args = self.get_args()
        in_args[-1] = -np.ones_like(in_args[-1])
        out = f(*in_args)
        assert np.allclose(0.0, out)

    def test_infer_shape(self):
        f, args, arg_vals = self.get_args()
        self._compile_and_check(args, self.op(*args), arg_vals, self.op_class)

    def test_grad(self):
        _, _, in_args = self.get_args()
        func = lambda *args: self.op(*args)[0]  # NOQA
        tt.verify_grad(func, in_args, rng=np.random)

    def test_pickle(self):
        f, _, in_args = self.get_args()
        data = pickle.dumps(self.op, -1)
        new_op = pickle.loads(data)
        assert np.allclose(f(*in_args), new_op(*in_args)[0].eval())


class TestRadiusFromOccArea(InferShapeTester):
    def setup_method(self):
        super().setup_method()
        self.op_class = RadiusFromOccArea
        self.op = RadiusFromOccArea()

    def get_args(self):
        delta = tt.vector()
        b = tt.vector()
        f = theano.function([delta, b], self.op(delta, b))

        delta_val = np.linspace(0.01, 0.99, 100)
        b_val = np.linspace(0.01, 1.5, len(delta_val))

        return f, [delta, b], [delta_val, b_val]

    def test_basic(self):
        f, t_args, v_args = self.get_args()
        r = f(*v_args)
        expect = -LimbDark()(
            np.array([1.0 / np.pi, 0.0]), v_args[1], r, np.ones_like(r)
        )[0].eval()
        assert np.allclose(expect, v_args[0])

    def test_infer_shape(self):
        f, args, arg_vals = self.get_args()
        self._compile_and_check(
            args, [self.op(*args)], arg_vals, self.op_class
        )

    def test_grad(self):
        func, t_args, v_args = self.get_args()
        g = theano.function(
            t_args, theano.grad(tt.sum(self.op(*t_args)), t_args)
        )(*v_args)
        eps = 1.234e-6
        for n in range(len(t_args)):
            v_args[n] += eps
            plus = func(*v_args)
            v_args[n] -= 2 * eps
            minus = func(*v_args)
            v_args[n] += eps

            est = 0.5 * (plus - minus) / eps
            assert np.allclose(est, g[n], atol=2 * eps)


def test_simple():
    np.random.seed(4502934)
    r = np.random.uniform(0, 1.0, 50)
    b = np.random.uniform(-1.2, 1.2, len(r))

    u = np.array([0.5, 0.3])
    cl = GetCl()(tt.as_tensor_variable(u)).eval()
    cl /= np.pi * (cl[0] + 2 * cl[1] / 3)
    f0 = LimbDark()(cl, b, r, np.ones_like(b))[0].eval()

    ld = SimpleLimbDark()
    assert np.allclose(ld.apply(b, r), 0.0)

    ld.set_u(u)
    assert np.allclose(ld.apply(b, r), f0)


def test_simple_pickle():
    np.random.seed(4502934)
    r = np.random.uniform(0, 1.0, 50)
    b = np.random.uniform(-1.2, 1.2, len(r))
    u = np.array([0.5, 0.3])

    ld = SimpleLimbDark()
    ld.set_u(u)
    f0 = ld.apply(b, r)

    data = pickle.dumps(ld, -1)
    ld2 = pickle.loads(data)
    assert np.allclose(ld2.apply(b, r), f0)
